@@ -31,6 +31,10 @@
 
 #include "version.h"
 
+#ifdef _MSC_VER
+#define strncasecmp _strnicmp
+#endif
+
 #define OUTPUT_URL      0  /* default */
 #define OUTPUT_SCHEME   1
 #define OUTPUT_USER     2
@@ -77,6 +81,8 @@ static const struct var variables[] = {
 #define ERROR_SET    5 /* a --set problem */
 #define ERROR_MEM    6 /* out of memory */
 #define ERROR_URL    7 /* could not get a URL out of the set components */
+#define ERROR_TRIM   8 /* a --trim problem */
+#define ERROR_BADURL 9 /* if --verify is set and the URL cannot parse */
 
 static void warnf(char *fmt, ...)
 {
@@ -103,18 +109,19 @@ static void errorf(int exit_code, char *fmt, ...)
 static void help(void)
 {
   int i;
-  fprintf(stderr, "Usage: [options] [URL]\n"
-          "  -h, --help                   - this help\n"
-          "  -v, --version                - show version\n"
-          " INPUT\n"
+  fprintf(stderr, "Usage: " PROGNAME " [options] [URL]\n"
+          "  --accept-space               - give in to this URL abuse\n"
           "  --append [component]=[data]  - append data to component\n"
+          "  -f, --url-file [file/-]      - read URLs from file or stdin\n"
+          "  -g, --get [{component}s]     - output URL component(s)\n"
+          "  -h, --help                   - this help\n"
+          " --json                        - output URL info as JSON\n"
           "  --redirect [URL]             - redirect the base URL to this\n"
           "  -s, --set [component]=[data] - set this component\n"
+          "  --trim [component]=[what]    - trim a component\n"
           "  --url [base URL]             - URL to start with\n"
-          "  -f, --url-file [file/-]      - read URLs from file or stdin\n"
-          " OUTPUT\n"
-          "  -g, --get [{component}s]     - output URL component(s)\n"
-          " --json                        - output URL info as JSON\n"
+          "  -v, --version                - show version\n"
+          " --verify                      - return error on (first) bad URL\n"
           " URL COMPONENTS:\n"
           "  "
     );
@@ -138,12 +145,18 @@ struct option {
   struct curl_slist *append_path;
   struct curl_slist *append_query;
   struct curl_slist *set_list;
+  struct curl_slist *trim_list;
   const char *redirect;
   const char *format;
   FILE *url;
   bool urlopen;
   bool jsonout;
+  bool verify;
+  bool accept_space;
   unsigned char output;
+
+  /* -- stats -- */
+  unsigned int urls;
 };
 
 static void urladd(struct option *o, const char *url)
@@ -229,6 +242,15 @@ static void setadd(struct option *o,
     o->set_list = n;
 }
 
+static void trimadd(struct option *o,
+                    const char *trim) /* [component]=[data] */
+{
+  struct curl_slist *n;
+  n = curl_slist_append(o->trim_list, trim);
+  if(n)
+    o->trim_list = n;
+}
+
 static bool checkoptarg(const char *str,
                         const char *given,
                         const char *arg)
@@ -276,6 +298,10 @@ static int getarg(struct option *op,
     op->redirect = arg;
     *usedarg = 1;
   }
+  else if(checkoptarg("--trim", flag, arg)) {
+    trimadd(op, arg);
+    *usedarg = 1;
+  }
   else if(checkoptarg("-g", flag, arg) ||
           checkoptarg("--get", flag, arg)) {
     if(op->format)
@@ -285,6 +311,10 @@ static int getarg(struct option *op,
   }
   else if(!strcmp("--json", flag))
     op->jsonout = true;
+  else if(!strcmp("--verify", flag))
+    op->verify = true;
+  else if(!strcmp("--accept-space", flag))
+    op->accept_space = true;
   else
     return 1;  /* unrecognized option */
   return 0;
@@ -326,7 +356,9 @@ static void get(struct option *op, CURLU *uh)
              !strncasecmp(ptr, variables[i].name, vlen)) {
             char *nurl;
             CURLUcode rc;
-            rc = curl_url_get(uh, variables[i].part, &nurl, CURLU_DEFAULT_PORT|
+            rc = curl_url_get(uh, variables[i].part, &nurl,
+                              CURLU_DEFAULT_PORT|
+                              CURLU_NO_DEFAULT_PORT|
                               (urldecode?CURLU_URLDECODE:0));
             switch(rc) {
             case CURLUE_OK:
@@ -404,7 +436,7 @@ static void set(CURLU *uh,
           if(varset[i])
             errorf(ERROR_SET, "A component can only be set once per URL (%s)",
                    variables[i].name);
-          curl_url_set(uh, variables[i].part, ptr+1,
+          curl_url_set(uh, variables[i].part, ptr[1] ? &ptr[1] : NULL,
                        CURLU_NON_SUPPORT_SCHEME|
                        (urlencode ? CURLU_URLENCODE : 0) );
           found = true;
@@ -470,19 +502,93 @@ static void json(struct option *o, CURLU *uh)
 {
   int i;
   (void)o;
-  fputs("{\n", stdout);
+  printf("%s  {\n", o->urls?",\n":"");
   for(i = 0; variables[i].name; i++) {
     char *nurl;
     CURLUcode rc = curl_url_get(uh, variables[i].part, &nurl,
                                 (i?CURLU_DEFAULT_PORT:0)|
                                 CURLU_URLDECODE);
     if(!rc) {
-      printf("  \"%s\": ", variables[i].name);
+      if(i)
+        fputs(",\n", stdout);
+      printf("    \"%s\": ", variables[i].name);
       jsonString(stdout, nurl, false);
-      fputs(",\n", stdout);
     }
   }
-  fputs("}\n", stdout);
+  fputs("\n  }", stdout);
+}
+
+/* --trim query="utm_*" */
+static void trim(CURLU *uh, struct option *o)
+{
+  struct curl_slist *node;
+  for(node = o->trim_list; node; node=node->next) {
+    char *instr = node->data;
+    if(strncasecmp(instr, "query", 5))
+      /* for now we can only trim query components */
+      errorf(ERROR_TRIM, "Unsupported trim component: %s", instr);
+    char *ptr = strchr(instr, '=');
+    if(ptr && (ptr > instr)) {
+      /* 'ptr' should be a fixed string or a pattern ending with an
+         asterisk */
+      size_t inslen;
+      bool pattern;
+      char *q;
+      CURLUcode rc;
+
+      ptr++; /* pass the = */
+      inslen = strlen(ptr);
+      pattern = ptr[inslen - 1] == '*';
+      if(pattern)
+        inslen--;
+      rc = curl_url_get(uh, CURLUPART_QUERY, &q, 0);
+      if(!rc && q) {
+        bool updated = 0;
+        char *newq = q;
+
+        do {
+          char *end = strchr(q, '&');
+          char *sep = strchr(q, '=');
+
+          if (! end) {
+            end = q + strlen(q);
+          }
+          if (! sep || sep > end) {
+            sep = end;
+          }
+
+          size_t qlen = sep - q;
+          if((pattern && (inslen <= qlen) &&
+              !strncasecmp(q, ptr, inslen)) ||
+              (!pattern && (inslen == qlen) &&
+              !strncasecmp(q, ptr, inslen))) {
+              /* this part should be stripped out */
+              if(end && *end) {
+                end++;
+                /* move the rest of the string to this position */
+                memmove(q, end, strlen(end)+1);
+              }
+              else {
+                /* don't leave a trailing ampersand */
+                if((q > newq) && (q[-1] == '&'))
+                  q--;
+                *q = 0;
+              }
+              updated = true;
+          }
+          else {
+            q = end;
+            if(!q)
+              break;
+            q++;
+          }
+        } while(*q);
+        if(updated)
+          rc = curl_url_set(uh, CURLUPART_QUERY, newq, 0);
+        curl_free(newq);
+      }
+    }
+  }
 }
 
 static void singleurl(struct option *o,
@@ -495,10 +601,13 @@ static void singleurl(struct option *o,
     if(url) {
       CURLUcode rc =
         curl_url_set(uh, CURLUPART_URL, url,
-                     CURLU_GUESS_SCHEME|CURLU_NON_SUPPORT_SCHEME);
+                     CURLU_GUESS_SCHEME|CURLU_NON_SUPPORT_SCHEME|
+                     (o->accept_space ?
+                      (CURLU_ALLOW_SPACE|CURLU_URLENCODE) : 0));
       if(rc) {
+        if(o->verify)
+          errorf(ERROR_BADURL, "%s [%s]", curl_url_strerror(rc), url);
         warnf("%s [%s]", curl_url_strerror(rc), url);
-        return;
       }
       else {
         if(o->redirect)
@@ -540,18 +649,24 @@ static void singleurl(struct option *o,
       char *oq;
       char *nq;
       /* extract the current query */
-      curl_url_get(uh, CURLUPART_QUERY, &oq, 0);
-
-      /* append the new segment */
-      nq = curl_maprintf("%s&%s", oq, aq);
-      if(nq) {
-        /* set the new query */
-        curl_url_set(uh, CURLUPART_QUERY, nq, 0);
+      if(!curl_url_get(uh, CURLUPART_QUERY, &oq, 0)) {
+        /* append the new segment */
+        nq = curl_maprintf("%s&%s", oq, aq);
+        if(nq)
+          /* set the new query */
+          curl_url_set(uh, CURLUPART_QUERY, nq, 0);
+        curl_free(nq);
+        curl_free(oq);
+      }
+      else {
+        /* no existing query, set the new one */
+        curl_url_set(uh, CURLUPART_QUERY, aq, 0);
       }
-      curl_free(nq);
-      curl_free(oq);
     }
 
+    /* trim parts */
+    trim(uh, o);
+
     if(o->jsonout)
       json(o, uh);
     else if(o->format) {
@@ -561,7 +676,7 @@ static void singleurl(struct option *o,
     else {
       /* default output is full URL */
       char *nurl = NULL;
-      if(!curl_url_get(uh, CURLUPART_URL, &nurl, 0)) {
+      if(!curl_url_get(uh, CURLUPART_URL, &nurl, CURLU_NO_DEFAULT_PORT)) {
         printf("%s\n", nurl);
         curl_free(nurl);
       }
@@ -569,6 +684,7 @@ static void singleurl(struct option *o,
         errorf(ERROR_URL, "not enough input for a URL");
       }
     }
+    o->urls++;
     curl_url_cleanup(uh);
 }
 
@@ -600,6 +716,9 @@ int main(int argc, const char **argv)
     }
   }
 
+  if(o.jsonout)
+    fputs("[\n", stdout);
+
   if(o.url) {
     /* this is a file to read URLs from */
     char buffer[4096]; /* arbitrary max */
@@ -632,9 +751,12 @@ int main(int argc, const char **argv)
         singleurl(&o, NULL);
     } while(node);
   }
+  if(o.jsonout)
+    fputs("\n]\n", stdout);
   /* we're done with libcurl, so clean it up */
   curl_slist_free_all(o.url_list);
   curl_slist_free_all(o.set_list);
+  curl_slist_free_all(o.trim_list);
   curl_slist_free_all(o.append_path);
   curl_slist_free_all(o.append_query);
   curl_global_cleanup();
